@@ -10,20 +10,20 @@
 
 /*
  * Copyright (C) 2001, 2004, 2011, 2012, 2013 the Free Software Foundation, Inc.
- * 
+ *
  * This file is part of GAWK, the GNU implementation of the
  * AWK Programming Language.
- * 
+ *
  * GAWK is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation; either version 3 of the License, or
  * (at your option) any later version.
- * 
+ *
  * GAWK is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
- * 
+ *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
@@ -49,7 +49,7 @@
 #define N_(msgid) msgid
 
 static const gawk_api_t *api;	/* for convenience macros to work */
-static awk_ext_id_t *ext_id;
+static awk_ext_id_t ext_id;
 static const char *ext_version = "ordchr extension: version 1.0";
 static awk_bool_t (*init_func)(void) = NULL;
 
@@ -58,24 +58,17 @@ int plugin_is_GPL_compatible;
 /*  do_ord --- return numeric value of first char of string */
 
 static awk_value_t *
-do_ord(int nargs, awk_value_t *result)
+do_ord(int nargs, awk_value_t *result, struct awk_ext_func *unused)
 {
 	awk_value_t str;
 	double ret = -1;
 
 	assert(result != NULL);
 
-	if (do_lint && nargs > 1)
-		lintwarn(ext_id, _("ord: called with too many arguments"));
-
 	if (get_argument(0, AWK_STRING, & str)) {
 		ret = str.str_value.str[0];
-	} else if (do_lint) {
-		if (nargs == 0)
-			lintwarn(ext_id, _("ord: called with no arguments"));
-		else
-			lintwarn(ext_id, _("ord: called with inappropriate argument(s)"));
-	}
+	} else if (do_lint)
+		lintwarn(ext_id, _("ord: called with inappropriate argument(s)"));
 
 	/* Set the return value */
 	return make_number(ret, result);
@@ -84,7 +77,7 @@ do_ord(int nargs, awk_value_t *result)
 /*  do_chr --- turn numeric value into a string */
 
 static awk_value_t *
-do_chr(int nargs, awk_value_t *result)
+do_chr(int nargs, awk_value_t *result, struct awk_ext_func *unused)
 {
 	awk_value_t num;
 	unsigned int ret = 0;
@@ -95,29 +88,22 @@ do_chr(int nargs, awk_value_t *result)
 
 	assert(result != NULL);
 
-	if (do_lint && nargs > 1)
-		lintwarn(ext_id, _("chr: called with too many arguments"));
-
 	if (get_argument(0, AWK_NUMBER, & num)) {
 		val = num.num_value;
 		ret = val;	/* convert to int */
 		ret &= 0xff;
 		str[0] = ret;
 		str[1] = '\0';
-	} else if (do_lint) {
-		if (nargs == 0)
-			lintwarn(ext_id, _("chr: called with no arguments"));
-		else
-			lintwarn(ext_id, _("chr: called with inappropriate argument(s)"));
-	}
+	} else if (do_lint)
+		lintwarn(ext_id, _("chr: called with inappropriate argument(s)"));
 
 	/* Set the return value */
 	return make_const_string(str, 1, result);
 }
 
 static awk_ext_func_t func_table[] = {
-	{ "ord", do_ord, 1 },
-	{ "chr", do_chr, 1 },
+	{ "ord", do_ord, 1, 1, awk_false, NULL },
+	{ "chr", do_chr, 1, 1, awk_false, NULL },
 };
 
 /* define the dl_load function using the boilerplate macro */
