@@ -3,7 +3,7 @@
  */
 
 /*
- * Copyright (C) 1986, 1988, 1989, 1991-2017 the Free Software Foundation, Inc.
+ * Copyright (C) 1986, 1988, 1989, 1991-2018 the Free Software Foundation, Inc.
  *
  * This file is part of GAWK, the GNU implementation of the
  * AWK Programming Language.
@@ -308,6 +308,9 @@ set_record(const char *buf, int cnt, const awk_fieldwidth_info_t *fw)
 	n->valref = 1;
 	n->type = Node_val;
 	n->stfmt = STFMT_UNUSED;
+#ifdef HAVE_MPFR
+	n->strndmode = MPFR_round_mode;
+#endif
 	n->flags = (STRING|STRCUR|USER_INPUT);	/* do not set MALLOC */
 	fields_arr[0] = n;
 	if (fw != api_fw) {
@@ -341,14 +344,20 @@ static void
 purge_record()
 {
 	int i;
-	NODE *n;
 
 	NF = -1;
 	for (i = 1; i <= parse_high_water; i++) {
-		assert((fields_arr[i]->flags & MALLOC) == 0
-			? fields_arr[i]->valref == 1
-			: true);
-		unref(fields_arr[i]);
+		NODE *n;
+		NODE *r = fields_arr[i];
+		if ((r->flags & MALLOC) == 0 && r->valref > 1) {
+			/* This can and does happen. We must copy the string! */
+			const char *save = r->stptr;
+			emalloc(r->stptr, char *, r->stlen + 1, "purge_record");
+			memcpy(r->stptr, save, r->stlen);
+			r->stptr[r->stlen] = '\0';
+			r->flags |= MALLOC;
+		}
+		unref(r);
 		getnode(n);
 		*n = *Null_field;
 		fields_arr[i] = n;
