@@ -1,5 +1,5 @@
 /* gawkmisc.c --- miscellaneous gawk routines that are OS specific.
- 
+
    Copyright (C) 1986, 1988, 1989, 1991 - 1998, 2001 - 2004, 2011
    the Free Software Foundation, Inc.
 
@@ -40,7 +40,7 @@ char *
 gawk_name(const char *filespec)
 {
 	char *p;
-    
+
 	/* "path/name" -> "name" */
 	p = strrchr(filespec, '/');
 	return (p == NULL ? (char *) filespec : p + 1);
@@ -127,7 +127,7 @@ optimal_bufsize(int fd, struct stat *stb)
 	 * guess. We use stdio's BUFSIZ, since that is what it was
 	 * meant for in the first place.
 	 */
-#ifdef HAVE_ST_BLKSIZE
+#ifdef HAVE_STRUCT_STAT_ST_BLKSIZE
 #define DEFBLKSIZE	(stb->st_blksize > 0 ? stb->st_blksize : BUFSIZ)
 #else
 #define DEFBLKSIZE	BUFSIZ
@@ -272,7 +272,7 @@ os_isatty(int fd)
 {
 	return isatty(fd);
 }
- 
+
 /* files_are_same --- return true if files are identical */
 
 int
@@ -285,21 +285,7 @@ files_are_same(char *path, SRCFILE *src)
 		&& st.st_ino == src->sbuf.st_ino);
 }
 
-#ifdef __CYGWIN__
 void
-cygwin_premain0(int argc, char **argv, struct per_process *myself)
+init_sockets(void)
 {
-	static struct __cygwin_perfile pf[] = {
-		{ "", O_RDONLY | O_TEXT },
-		/*{ "", O_WRONLY | O_BINARY },*/
-		{ NULL, 0 }
-	};
-	cygwin_internal(CW_PERFILE, pf);
 }
-
-void
-cygwin_premain2(int argc, char **argv, struct per_process *myself)
-{
-	setmode(fileno (stdin), O_TEXT);
-}
-#endif
